# Experimental DRAKON to XML convertor for DRAKON Editor 1.15
# Usage: tclsh drn_to_xml.tcl input.drn output.xml
# Author: Stepan Mitkin stipan.mitkin@gmail.com
# http://drakon-editor.sourceforge.net/
# Date: 27 March 2013
# License: The content of this file is PUBLIC DOMAIN.

package require sqlite3


proc main { arguments } {
	set db "input-database"

	if { [ llength $arguments ] != 2 } {
		puts "Usage: tclsh drn_to_xml.tcl input.drn output.xml"
		exit 1
	}

	set drn_filename [ lindex $arguments 0 ]
	set xml_filename [ lindex $arguments 1 ]

	if { $drn_filename == $xml_filename } {
		puts "Input and output filenames are the same."
		exit 1
	}
	
	if { ![ file exists $drn_filename ] } {
		puts "Input file does not exist: $drn_filename"
		exit 1
	}

	sqlite3 $db $drn_filename -readonly yes
	set version [ get_version $db ]
	
	set xml_file [ open $xml_filename "wb" ]
	
	puts $xml_file "<?xml version=\"1.0\" encoding=\"UTF-8\"?>"
	
	begin_tag $xml_file "drakon"
	attribute $xml_file "version" $version
	end_tag $xml_file
	
	save_diagrams $db $xml_file
	save_tree $db $xml_file
	
	
	close_tag $xml_file "drakon"
	
	close $xml_file
	$db close
}

proc get_version { db } {
	set version [ $db onecolumn {
		select value
		from info
		where key = 'version'
	} ]

	set start_version [ $db onecolumn {
		select value
		from info
		where key = 'start_version'
	} ]
	
	return "$start_version.$version"
}

proc begin_tag { file name } {
	puts $file "<$name"
}

proc end_tag { file } {
	puts $file ">"
}

proc attribute { file name value } {
	set escaped [ escape_xml $value ]
	set encoded [ encoding convertto utf-8 $escaped ]
	puts $file "$name=\"$encoded\""
}

proc escape_xml { text } {
	set xml_escapes {
		"\""	&quot;
		"\'"	&apos;
		<		&lt;
		>		&gt;
		&		&amp;
	}
	return [ string map $xml_escapes $text ]	
}

proc open_tag { file name } {
	puts $file "<$name>"
}

proc close_tag { file name } {
	puts $file "</$name>"
}

proc save_diagrams { db file } {
	open_tag $file "diagrams"
	
	$db eval {
		select *
		from diagrams
		order by diagram_id
	} row {
		set fields [ array get row ]
		
		begin_tag $file "diagram"
		save_attribute $file $fields "diagram_id"
		end_tag $file
		
		save_cdata $file $fields "name"
		save_cdata $file $fields "description"
		
		save_items $db $file $row(diagram_id)
		
		close_tag $file "diagram"
	}
	close_tag $file "diagrams"
}

proc save_items { db file diagram_id } {
	open_tag $file "items"
	
	$db eval {
		select *
		from items
		where diagram_id = :diagram_id
		order by item_id
	} row {
		set fields [ array get row ]
		
		begin_tag $file "item"
		save_attribute $file $fields "item_id"
		save_attribute $file $fields "type"
		save_attribute $file $fields "x"
		save_attribute $file $fields "y"
		save_attribute $file $fields "w"
		save_attribute $file $fields "h"
		save_attribute $file $fields "a"
		save_attribute $file $fields "b"
		end_tag $file
		
		save_cdata $file $fields "text"
		close_tag $file "item"
	}
	close_tag $file "items"
}

proc save_attribute { file fields name } {
	set value [ dict get $fields $name ]	
	attribute $file $name $value
}

proc save_cdata { file fields name } {
	set value [ dict get $fields $name ]
	if { $value != "" } {
		set escaped [ cdata_escape $value ]
		set encoded [ encoding convertto utf-8 $escaped ]
		puts $file "<$name><!\[CDATA\["
		puts $file $encoded
		puts $file "\]\]></$name>"
	}
}

proc cdata_escape { text } {
	return [ string map { "\]\]>" "\]\]\]\]><!\[CDATA\[>" } $text ]
}

proc save_tree { db file } {
	open_tag $file "tree_nodes"
	save_tree_nodes $db $file 0
	close_tag $file "tree_nodes"
}

proc save_tree_nodes { db file parent } {
	open_tag $file "tree_node"
	$db eval {
		select *
		from tree_nodes
		where parent = :parent
	} row {
		set fields [ array get row ]
		
		begin_tag $file "item"
		save_attribute $file $fields "node_id"
		save_attribute $file $fields "type"
		save_attribute $file $fields "name"
		save_attribute $file $fields "diagram_id"
		end_tag $file
		
		save_tree_nodes $db $file $row(node_id)
		close_tag $file "item"		
	}
	close_tag $file "tree_node"
}

main $argv